import ipaddress
from uuid import uuid4

from orchestrator.db.models import ProductTable, SubscriptionTable
# noinspection PyProtectedMember
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, choice_list
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State
from orchestrator.types import SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.device import Device
from gso.products.product_types.iptrunk import IptrunkInactive, \
    IptrunkProvisioning
from gso.services import provisioning_proxy
from gso.services.provisioning_proxy import confirm_pp_results, \
    await_pp_results


def device_selector(choice_value: str) -> list:
    device_subscriptions = {}
    for device_id, device_description in (
            SubscriptionTable.query.join(ProductTable)
                                   .filter(
                ProductTable.product_type == 'Device',
                SubscriptionTable.status == 'active',
            )
            .with_entities(SubscriptionTable.subscription_id,
                           SubscriptionTable.description)
            .all()
    ):
        device_subscriptions[str(device_id)] = device_description

    # noinspection PyTypeChecker
    return choice_list(
        Choice(choice_value, zip(device_subscriptions.keys(),
                                 device_subscriptions.items())),  # type:ignore
        min_items=1,
        max_items=1,
    )


def initial_input_form_generator(product_name: str) -> FormGenerator:
    class CreateIptrunkForm(FormPage):
        class Config:
            title = product_name

        geant_s_sid: str
        iptrunk_description: str
        iptrunk_type: IptrunkType
        iptrunk_speed: str  # This should be an enum: 1/10/100/400
        iptrunk_minimum_links: int

        iptrunk_sideA_node_id: device_selector(
            choice_value='DeviceEnumA')  # noqa: F821
        iptrunk_sideA_ae_iface: str
        iptrunk_sideA_ae_geant_a_sid: str
        iptrunk_sideA_ae_members: list[str]
        iptrunk_sideA_ae_members_descriptions: list[str]

        iptrunk_sideB_node_id: device_selector(
            choice_value='DeviceEnumB')  # noqa: F821
        iptrunk_sideB_ae_iface: str
        iptrunk_sideB_ae_geant_a_sid: str
        iptrunk_sideB_ae_members: list[str]
        iptrunk_sideB_ae_members_descriptions: list[str]
        # TODO: we need additional validation:
        # * sideA fqdn must be different from sideB fqdn
        # * the length of iptrunk_sideA_ae_members should
        # be the same as iptrunk_sideB_ae_members
        # * interface names must be validated

    user_input = yield CreateIptrunkForm

    return user_input.dict()


@step('Create subscription')
def create_subscription(product: UUIDstr) -> State:
    subscription = IptrunkInactive.from_product_id(product, uuid4())

    return {
        'subscription': subscription,
        'subscription_id': subscription.subscription_id,
    }


@step('Get information from IPAM')
def get_info_from_ipam(subscription: IptrunkInactive) -> State:
    # TODO: get info about how these should be generated
    subscription.iptrunk.iptrunk_ipv4_network \
        = ipaddress.ip_network('192.168.255.0/31')
    subscription.iptrunk.iptrunk_ipv6_network \
        = ipaddress.ip_network('fc00:798:255::150/126')
    return {'subscription': subscription}


@step('Initialize subscription')
def initialize_subscription(
        subscription: IptrunkInactive,
        geant_s_sid: str,
        iptrunk_type: IptrunkType,
        iptrunk_description: str,
        iptrunk_speed: str,
        iptrunk_minimum_links: int,
        iptrunk_sideA_node_id: str,
        iptrunk_sideA_ae_iface: str,
        iptrunk_sideA_ae_geant_a_sid: str,
        iptrunk_sideA_ae_members: list[str],
        iptrunk_sideA_ae_members_descriptions: list[str],
        iptrunk_sideB_node_id: str,
        iptrunk_sideB_ae_iface: str,
        iptrunk_sideB_ae_geant_a_sid: str,
        iptrunk_sideB_ae_members: list[str],
        iptrunk_sideB_ae_members_descriptions: list[str]
) -> State:
    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sideA_node = Device.from_subscription(
        iptrunk_sideA_node_id[0]).device
    subscription.iptrunk.iptrunk_sideA_ae_iface = iptrunk_sideA_ae_iface
    subscription.iptrunk.iptrunk_sideA_ae_geant_a_sid \
        = iptrunk_sideA_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideA_ae_members = iptrunk_sideA_ae_members
    subscription.iptrunk.iptrunk_sideA_ae_members_description \
        = iptrunk_sideA_ae_members_descriptions

    subscription.iptrunk.iptrunk_sideB_node = Device.from_subscription(
        iptrunk_sideB_node_id[0]).device
    subscription.iptrunk.iptrunk_sideB_ae_iface \
        = iptrunk_sideB_ae_iface
    subscription.iptrunk.iptrunk_sideB_ae_geant_a_sid \
        = iptrunk_sideB_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideB_ae_members = iptrunk_sideB_ae_members
    subscription.iptrunk.iptrunk_sideB_ae_members_description \
        = iptrunk_sideB_ae_members_descriptions

    subscription.description = f'IP trunk, geant_s_sid:{geant_s_sid}'
    subscription = IptrunkProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING
    )

    return {'subscription': subscription}


@step('Provision IP trunk interface [DRY RUN]')
def provision_ip_trunk_iface_dry(subscription: IptrunkProvisioning,
                                 process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'trunk_interface')

    return {'subscription': subscription}


@step('Provision IP trunk interface [FOR REAL]')
def provision_ip_trunk_iface_real(subscription: IptrunkProvisioning,
                                  process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'trunk_interface', False)

    return {'subscription': subscription}


@step('Provision IP trunk ISIS interface [DRY RUN]')
def provision_ip_trunk_isis_iface_dry(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'isis_interface')

    return {'subscription': subscription}


@step('Provision IP trunk ISIS interface [FOR REAL]')
def provision_ip_trunk_isis_iface_real(subscription: IptrunkProvisioning,
                                       process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'isis_interface', False)

    return {'subscription': subscription}


@step('Provision IP trunk LDP interface [DRY RUN]')
def provision_ip_trunk_ldp_iface_dry(subscription: IptrunkProvisioning,
                                     process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'ldp_interface')

    return {'subscription': subscription}


@step('Provision IP trunk LDP interface [FOR REAL]')
def provision_ip_trunk_ldp_iface_real(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'ldp_interface', False)

    return {'subscription': subscription}


@step('Provision IP trunk LLDP interface [DRY RUN]')
def provision_ip_trunk_lldp_iface_dry(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'lldp_interface')

    return {'subscription': subscription}


@step('Provision IP trunk LLDP interface [FOR REAL]')
def provision_ip_trunk_lldp_iface_real(subscription: IptrunkProvisioning,
                                       process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'lldp_interface', False)

    return {'subscription': subscription}


@workflow(
    'Create IP trunk',
    initial_input_form=wrap_create_initial_input_form(
        initial_input_form_generator),
    target=Target.CREATE,
)
def create_iptrunk():
    return (
            init
            >> create_subscription
            >> store_process_subscription(Target.CREATE)
            >> get_info_from_ipam
            >> initialize_subscription
            >> provision_ip_trunk_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_isis_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_isis_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_ldp_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_ldp_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_lldp_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_lldp_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> set_status(SubscriptionLifecycle.ACTIVE)
            >> resync
            >> done
    )
